;; Generated package description from osm.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "osm" "2.3.0.20260519.102419" "OpenStreetMap viewer" '((emacs "29.1") (compat "31")) :commit "a9e76ce08fb86e40a9f3efc42b38045c9b58fa95" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("network" "multimedia" "hypermedia" "mouse") :url "https://github.com/minad/osm")
